"""Module implementing unions of basic bounds."""

import itertools
import numpy as np

from sklearn.mixture import GaussianMixture
from scipy.optimize import minimize
from scipy.special import logsumexp
from scipy.stats import multivariate_normal

from .basic import UnitCube, Ellipsoid, UnitCubeEllipsoidMixture


def ellipsoids_overlap(ellipsoids):
    """Determine if ellipsoids overlap.

    This functions is based on ieeexplore.ieee.org/document/6289830.

    Parameters
    ----------
    ellipsoids : list
        List of ellipsoids.

    Returns
    -------
    overlapping : bool
        True, if there is overlap between ellipoids and False otherwise.

    """
    c = [ellipsoid.c for ellipsoid in ellipsoids]
    A_inv = [np.linalg.inv(ellipsoid.A) for ellipsoid in ellipsoids]

    for i_1, i_2 in itertools.combinations(range(len(c)), 2):
        d = c[i_1] - c[i_2]
        def k(s): return (1 - np.dot(np.dot(
            d, np.linalg.inv(A_inv[i_1] / (1 - s) + A_inv[i_2] / s)), d))
        if minimize(k, 0.5, bounds=[(1e-9, 1-1e-9)]).fun > 0:
            return True

    return False


class Union():
    r"""Union of multiple ellipsoids or unit cube-ellipsoid mixtures.

    Attributes
    ----------
    n_dim : int
        Number of dimensions.
    enlarge_per_dim : float
        Along each dimension, ellipsoids enlarged by this factor.
    n_points_min : int or None
        The minimum number of points each ellipsoid should have. Effectively,
        ellipsoids with less than twice that number will not be split further.
    cube : UnitCube or None
        If not None, the bound is defined as the overlap with the unit cube.
    points_bounds : list
        The points used to create the individual bounds. Used to split bounds
        further.
    bounds : list
        List of individual bounds.
    log_v_all : numpy.ndarray
        Natural log of the volume of each individual bound.
    block : numpy.ndarray
        List indicating whether a bound should not be split further because
        the volume would not decrease or the number of points is too low.
    points : numpy.ndarray
        Points that a call to `sample` will return next.
    n_sample : int
        Number of points sampled from all bounds.
    n_reject : int
        Number of points rejected due to overlap.
    rng : numpy.random.Generator
        Determines random number generation.

    """

    @classmethod
    def compute(cls, points, enlarge_per_dim=1.1, n_points_min=None,
                unit=True, bound_class=Ellipsoid, rng=None):
        """Compute the bound.

        Upon creation, the bound consists of a single individual bound.

        Parameters
        ----------
        points : numpy.ndarray with shape (n_points, n_dim)
            A 2-D array where each row represents a point.
        enlarge_per_dim : float, optional
            Along each dimension, the ellipsoid is enlarged by this factor.
            Default is 1.1.
        n_points_min : int or None, optional
            The minimum number of points each ellipsoid should have.
            Effectively, ellipsoids with less than twice that number will not
            be split further. If None, uses `n_points_min = n_dim + 1`. Default
            is None.
        unit : bool, optional
            If the bound is restricted to the overlap with the unit cube.
        bound_class : class, optional
            Type of the individual bounds, i.e. ellipsoids or unit
            cube-ellipsoid mixtures.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Raises
        ------
        ValueError
            If `n_points_min` is smaller than the number of dimensions plus
            one.

        Returns
        -------
        bound : Union
            The bound.

        """
        bound = cls()
        bound.n_dim = points.shape[1]
        bound.enlarge_per_dim = enlarge_per_dim

        if n_points_min is None:
            bound.n_points_min = bound.n_dim + 1
        else:
            if n_points_min < bound.n_dim + 1:
                raise ValueError('The number of points per bound must be ' +
                                 'larger than the number of dimensions.')
            bound.n_points_min = n_points_min

        if not unit:
            bound.cube = None
        else:
            bound.cube = UnitCube.compute(
                bound.n_dim, rng=rng)

        bound.points_bounds = [points]
        bound.bounds = [bound_class.compute(
            points, enlarge_per_dim=enlarge_per_dim,
            rng=rng)]
        bound.log_v_all = np.array([bound.bounds[0].log_v])
        bound.block = np.atleast_1d(len(points) < 2 * bound.n_points_min)

        bound.points = np.zeros((0, points.shape[1]))
        bound.n_sample = 0
        bound.n_reject = 0

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        return bound

    def split(self, allow_overlap=True):
        """Split the largest bound in the union.

        Parameters
        ----------
        allow_overlap : bool, optional
            Whether to allow splitting the largest bound if doing so creates
            overlaps between bounds. Cannot be False if the individual bounds
            are cube-ellipsoid mixtures. Default is True.

        Raises
        ------
        ValueError
            If `allow_overlap` is False and the individual bounds are
            cube-ellipsoid mixtures.

        Returns
        -------
        success : bool
            Whether it was possible to split any bound.

        """
        if not allow_overlap and not isinstance(self.bounds[0], Ellipsoid):
            raise ValueError("'allow_overlap' can only be False if " +
                             "bounds are ellipsoids.")

        if not np.any(~self.block):
            return False

        index = np.argmax(np.where(~self.block, self.log_v_all, -np.inf))
        points = self.bounds[index].transform(self.points_bounds[index])

        gmm = GaussianMixture(
            n_components=2, n_init=10,
            random_state=self.rng.integers(2**32 - 1)).fit(points)
        p = np.vstack([multivariate_normal.logpdf(
            points, mean=gmm.means_[i], cov=gmm.covariances_[i]) +
            np.log(gmm.weights_[i]) for i in range(2)]).T

        labels = np.argmax(p, axis=1)
        # If one of the clusters has less than n_points_min members, re-assign
        # the most likely members from the larger cluster to the smaller one.
        if not np.all(np.bincount(labels) >= self.n_points_min):
            label = np.argmin(np.bincount(labels))
            labels[np.argsort(-p[:, label])[:self.n_points_min]] = label

        new_bounds = []
        points = self.points_bounds[index]
        for label in [0, 1]:
            new_bounds.append(type(self.bounds[0]).compute(
                points[labels == label], enlarge_per_dim=self.enlarge_per_dim,
                rng=self.rng))

        if not allow_overlap and ellipsoids_overlap(
                self.bounds[:index] + self.bounds[index+1:] + new_bounds):
            return False

        if (logsumexp([new_bounds[0].log_v, new_bounds[1].log_v]) >
                self.bounds[index].log_v):
            self.block[index] = True
            return self.split(allow_overlap=allow_overlap)

        self.points_bounds.pop(index)
        self.points_bounds.append(points[labels == 0])
        self.points_bounds.append(points[labels == 1])
        self.bounds.pop(index)
        self.bounds = self.bounds + new_bounds
        self.log_v_all = np.array([bound.log_v for bound in self.bounds])
        self.block = np.concatenate(
            (np.delete(self.block, index),
             [len(self.points_bounds[-2]) < 2 * self.n_points_min,
              len(self.points_bounds[-1]) < 2 * self.n_points_min]))

        # Reset the sampling.
        self.reset()

        return True

    def trim(self, threshold=1e3):
        """Drop the lowest-density bound, if possible.

        Density is defined as the ratio of each bound's number of points to its
        volume.

        Parameters
        ----------
        threshold : float, optional
            Only drop the lowest-density bound if it has a density at least
            `threshold` times lower than the median of all other bounds.

        Returns
        -------
        success : bool
            Whether it was possible to drop a bound. Will always return False
            if there is only one bound in the union.

        """
        if len(self.bounds) == 1:
            return False

        log_n = np.array([np.log(len(points)) for points in
                          self.points_bounds])
        log_v = np.array([bound.log_v for bound in self.bounds])
        log_r = log_n - log_v
        index = np.argmin(log_r)

        if log_r[index] - np.median(np.delete(log_r, index)) < -np.log(
                threshold):
            self.points_bounds.pop(index)
            self.bounds.pop(index)
            self.log_v_all = np.array([bound.log_v for bound in self.bounds])
            self.reset()
            return True
        else:
            return False

    def contains(self, points):
        """Check whether points are contained in the bound.

        Parameters
        ----------
        points : numpy.ndarray
            A 1-D or 2-D array containing single point or a collection of
            points. If more than one-dimensional, each row represents a point.

        Returns
        -------
        in_bound : bool or numpy.ndarray
            Bool or array of bools describing for each point whether it is
            contained in the bound.

        """
        in_bound = np.any([bound.contains(points) for bound in self.bounds],
                          axis=0)
        if self.cube is not None:
            in_bound = in_bound & self.cube.contains(points)
        return in_bound

    def sample(self, n_points=100):
        """Sample points from the bound.

        Parameters
        ----------
        n_points : int, optional
            How many points to draw. Default is 100.

        Returns
        -------
        points : numpy.ndarray
            Points as two-dimensional array of shape (n_points, n_dim).

        """
        while len(self.points) < n_points:
            n_sample = 1000

            p = np.exp(np.array(self.log_v_all) - logsumexp(self.log_v_all))
            n_per_bound = self.rng.multinomial(n_sample, p)

            points = np.vstack([bound.sample(n) for bound, n in
                                zip(self.bounds, n_per_bound)])
            if self.cube is not None:
                points = points[self.cube.contains(points)]
            self.rng.shuffle(points)
            n_bound = np.sum([bound.contains(points) for bound in self.bounds],
                             axis=0)
            p = 1 - 1.0 / n_bound
            points = points[self.rng.random(size=len(points)) > p]
            self.points = np.vstack([self.points, points])

            self.n_sample += n_sample
            self.n_reject += n_sample - len(points)

        points = self.points[:n_points]
        self.points = self.points[n_points:]
        return points

    @property
    def log_v(self):
        """Return the natural log of the volume of the bound.

        Returns
        -------
        log_v : float
            The natural log of the volume.

        """
        if self.n_sample == 0:
            self.sample()

        return logsumexp(self.log_v_all) + np.log(
            1.0 - self.n_reject / self.n_sample)

    def write(self, group):
        """Write the bound to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['type'] = 'MultiEllipsoid'
        for key in ['n_dim', 'log_v_all', 'enlarge_per_dim', 'n_points_min',
                    'n_sample', 'n_reject']:
            group.attrs[key] = getattr(self, key)

        group.attrs['unit'] = self.cube is not None
        if self.cube is not None:
            self.cube.write(group.create_group('cube'))

        group.attrs['bound_class'] = self.bounds[0].__class__.__name__
        for i, bound in enumerate(self.bounds):
            bound.write(group.create_group('bound_{}'.format(i)))

        for i, points in enumerate(self.points_bounds):
            group.create_dataset('points_bound_{}'.format(i), data=points)
        group.create_dataset('points', data=self.points,
                             maxshape=(None, self.n_dim))

    def update(self, group):
        """Update bound information previously written to an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.

        """
        group.attrs['n_sample'] = self.n_sample
        group.attrs['n_reject'] = self.n_reject
        group['points'].resize(self.points.shape)
        group['points'][...] = self.points

    @classmethod
    def read(cls, group, rng=None):
        """Read the bound from an HDF5 group.

        Parameters
        ----------
        group : h5py.Group
            HDF5 group to write to.
        rng : None or numpy.random.Generator, optional
            Determines random number generation. Default is None.

        Returns
        -------
        bound : MultiEllipsoid
            The bound.

        """
        bound = cls()

        if rng is None:
            bound.rng = np.random.default_rng()
        else:
            bound.rng = rng

        for key in ['n_dim', 'log_v_all', 'enlarge_per_dim', 'n_points_min',
                    'n_sample', 'n_reject']:
            setattr(bound, key, group.attrs[key])

        if group.attrs['unit']:
            bound.cube = UnitCube.read(group['cube'], rng=bound.rng)

        if group.attrs['bound_class'] == 'Ellipsoid':
            bound_class = Ellipsoid
        else:
            bound_class = UnitCubeEllipsoidMixture

        bound.bounds = [bound_class.read(
            group['bound_{}'.format(i)], rng=bound.rng)
            for i in range(len(bound.log_v_all))]
        bound.points_bounds = [np.array(group['points_bound_{}'.format(i)]) for
                               i in range(len(bound.log_v_all))]
        bound.points = np.array(group['points'])

        return bound

    def reset(self, rng=None):
        """Reset random number generation and any progress, if applicable.

        Parameters
        ----------
        rng : None or numpy.random.Generator, optional
            Determines random number generation. If None, random number
            generation is not reset. Default is None.

        """
        self.points = np.zeros((0, self.n_dim))
        self.n_sample = 0
        self.n_reject = 0

        if rng is not None:
            self.rng = rng
            if self.cube is not None:
                self.cube.reset(rng)
            for bound in self.bounds:
                bound.reset(rng)



import numpy as np
import pytest
import warnings

from scipy.stats import multivariate_normal, norm

from nautilus import Prior, Sampler


@pytest.mark.parametrize("n_networks", [0, 1, 2])
@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("pass_dict", [True, False])
def test_sampler_basic(n_networks, vectorized, pass_dict):
    # Test that the sampler does not crash.

    if pass_dict:
        prior = Prior()
        prior.add_parameter('a')
        prior.add_parameter('b')
    else:
        def prior(x):
            return x

    def likelihood(x):
        if pass_dict:
            x = np.squeeze(np.column_stack([x['a'], x['b']]))
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=n_networks,
        vectorized=vectorized, pass_dict=pass_dict, n_live=200)
    sampler.run(n_like_max=600, verbose=True)
    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=n_networks,
        vectorized=vectorized, pass_dict=None, n_live=200)
    sampler.run(n_like_max=600, verbose=True)
    points, log_w, log_l = sampler.posterior()
    points, log_w, log_l = sampler.posterior(return_as_dict=pass_dict)
    points, log_w, log_l = sampler.posterior(
        return_as_dict=pass_dict, equal_weight=True)
    assert sampler.n_eff > 0
    sampler.log_z
    assert sampler.eta > 0
    assert sampler.eta < 1


def test_sampler_errors_and_warnings():
    # Test that the sampler correctly raises errors and warnings.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    with pytest.raises(ValueError):
        Sampler(prior, likelihood)

    with pytest.raises(ValueError):
        Sampler(prior, likelihood, 1)

    sampler = Sampler(prior, likelihood, 2, n_live=300)
    sampler.run(n_like_max=1000, verbose=True)

    with pytest.warns(DeprecationWarning):
        sampler.evidence()
    with pytest.warns(DeprecationWarning):
        sampler.effective_sample_size()
    with pytest.warns(DeprecationWarning):
        sampler.asymptotic_sampling_efficiency()
    with pytest.raises(ValueError):
        sampler.posterior(return_blobs=True)


@pytest.mark.parametrize("discard_exploration_start", [True, False])
@pytest.mark.parametrize("discard_exploration_end", [True, False, 1])
def test_sampler_switch_exploration(
        discard_exploration_start, discard_exploration_end):
    # Test that we can switch later whether to discard the exploration phase.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=1, vectorized=True, n_live=500)
    sampler.run(f_live=0.45, n_eff=10000, verbose=True,
                discard_exploration=discard_exploration_start)
    assert sampler.discard_exploration == discard_exploration_start
    points, log_w, log_l = sampler.posterior()
    n_start = len(points)
    log_z_start = sampler.log_z

    if not isinstance(discard_exploration_end, bool):
        with pytest.raises(ValueError):
            sampler.discard_exploration = discard_exploration_end
        return

    sampler.discard_exploration = discard_exploration_end
    points, log_w, log_l = sampler.posterior()
    n_end = len(points)
    log_z_end = sampler.log_z

    assert ((discard_exploration_start == discard_exploration_end) ==
            (n_start == n_end))
    assert ((discard_exploration_start == discard_exploration_end) ==
            (log_z_start == log_z_end))


@pytest.mark.parametrize("custom_prior", [True, False])
@pytest.mark.parametrize("vectorized", [True, False])
@pytest.mark.parametrize("pass_dict", [True, False])
def test_sampler_prior(custom_prior, vectorized, pass_dict):
    # Test that the sampler can handle all prior defintions.

    if custom_prior:
        if pass_dict:
            def prior(x):
                return dict(a=x[..., 0], b=x[..., 1])
        else:
            def prior(x):
                return x
    else:
        prior = Prior()
        prior.add_parameter('a')
        prior.add_parameter('b')

    def likelihood(x):
        if isinstance(x, dict):
            x = np.squeeze(np.column_stack([x['a'], x['b']]))
        return -np.linalg.norm(x - 0.5, axis=-1) * 0.001

    sampler = Sampler(
        prior, likelihood, n_dim=2, n_networks=1, vectorized=vectorized,
        pass_dict=pass_dict, n_live=500)
    sampler.run(f_live=0.45, n_eff=0, verbose=True)
    points, log_w, log_l = sampler.posterior(return_as_dict=pass_dict)
    if custom_prior and pass_dict:
        with pytest.raises(ValueError):
            points, log_w, log_l = sampler.posterior(return_as_dict=False)


@pytest.mark.parametrize("n_networks", [0, 4])
@pytest.mark.parametrize("discard_exploration", [True, False])
def test_sampler_accuracy(n_networks, discard_exploration):
    # Do a basic tests of the sampler accuracy.

    n_dim = 2
    mean = np.repeat(0.5, n_dim)
    cov = np.eye(n_dim) * 0.01

    def prior(x):
        return x

    def likelihood(x):
        return multivariate_normal.logpdf(x, mean=mean, cov=cov)

    sampler = Sampler(prior, likelihood, n_dim=n_dim, n_live=500,
                      n_networks=n_networks, seed=0)
    sampler.run(discard_exploration=discard_exploration, f_live=0.1,
                verbose=True)

    assert np.abs(sampler.log_z) < 0.05

    for equal_weight in [True, False]:
        points, log_w, log_l = sampler.posterior(equal_weight=equal_weight)
        weights = np.exp(log_w)

        mean_sampler = np.average(points, weights=weights, axis=0)
        cov_sampler = np.cov(points, aweights=weights, rowvar=False)
        assert np.all(np.isclose(mean_sampler, mean, atol=0.01, rtol=0))
        assert np.all(np.isclose(cov_sampler, cov, atol=0.001, rtol=0))

    # This is a simple problem so the bounds should be perfectly nested, i.e.
    # bound i+1 is fully contained within bound i.
    shell_bound_occupation = sampler.shell_bound_occupation()
    assert np.all(shell_bound_occupation ==
                  np.tril(np.ones_like(shell_bound_occupation)))


def test_sampler_enlarge_per_dim():
    # Test that the enlarge_per_dim keyword is passed correctly. In this
    # example, we choose an enlargment factor so big that every bound should be
    # the same if we don't use neural networks. Thus, after the exploration
    # phase, all but one bound should be left.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5)**2 * 0.001

    sampler = Sampler(prior, likelihood, n_dim=2, enlarge_per_dim=100,
                      n_networks=0, seed=0)
    sampler.run(f_live=0.1, n_eff=0, verbose=True)

    # The effective sample size should be very close to the number of calls
    # since the likelihood is extremely flat.
    assert np.isclose(sampler.n_like, sampler.n_eff, rtol=0, atol=1)
    # Only one bound should be left.
    assert len(sampler.bounds) == 1
    # Check evidence accuracy.
    assert np.isclose(sampler.log_z, -4 * 0.5**3 / 3 * 0.001, rtol=0,
                      atol=1e-4)


def test_sampler_empty_shells():
    # Test that the sampler can correctly deal with shells that are empty at
    # the end of the run. This is a somewhat contrived example that should
    # result in empty shells since each one should only have one sample, on
    # average.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5)**2 * 0.001

    sampler = Sampler(prior, likelihood, n_dim=2,
                      n_networks=0, seed=0, n_update=1, n_live=10, n_batch=1)
    sampler.run(f_live=1e-3, n_eff=0, verbose=True)


def test_sampler_n_like_max():
    # Test that the sampler correctly stops when hitting the maximum number of
    # likelihood calls and can resume afterwards.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5)**2 * 0.001

    sampler_a = Sampler(prior, likelihood, n_dim=2, n_networks=0, seed=0)
    sampler_b = Sampler(prior, likelihood, n_dim=2, n_networks=0, seed=0)

    sampler_a.run(verbose=True)
    for n_like_max in range(sampler_a.n_like + 1):
        success = sampler_b.run(n_like_max=n_like_max, verbose=True)
        assert sampler_b.n_like <= n_like_max + sampler_b.n_batch
        assert not success if sampler_a.n_like != sampler_b.n_like else success

    assert sampler_a.log_z == sampler_b.log_z
    assert sampler_a.n_eff == sampler_b.n_eff


def test_sampler_timeout():
    # Test that the sampler correctly stops when reaching the timeout limits.

    def prior(x):
        return x

    def likelihood(x):
        return -np.linalg.norm(x - 0.5)**2 * 0.001

    sampler = Sampler(prior, likelihood, n_dim=10, n_networks=0, seed=0)
    # The sampler shouldn't finish within 1 second.
    success = sampler.run(verbose=True, timeout=1)
    assert not success

    # We should be able to continue afterwards.
    success = sampler.run(verbose=True, timeout=5)


def test_sampler_funnel():
    # Test the sampler on a funnel distribution. This is a great, challenging
    # distribution. Also, the nature of the likelihood leads to nautilus
    # boundaries that are not strictly nested, unlike for simpler
    # distributions.

    def prior(x):
        return x

    def likelihood(x):
        return (norm.logpdf(x[0], loc=0.5, scale=0.1) +
                norm.logpdf(x[1], loc=0.5, scale=np.exp(20 * (x[0] - 0.5)) /
                            100))

    # Determine the true evidence. The likelihood is normalized so it should be
    # close to 1 minus a correction for the fraction of points falling outside
    # the unit cube.
    np.random.seed(0)
    x_0 = np.random.normal(loc=0.5, scale=0.1, size=1000000)
    x_1 = np.random.normal(loc=0.5, scale=np.exp(20 * (x_0 - 0.5)) / 100)
    log_z_true = np.log(np.mean((x_0 > 0) & (x_0 < 1) & (x_1 > 0) & (x_1 < 1)))

    sampler = Sampler(prior, likelihood, n_dim=2, n_networks=1, seed=0)
    sampler.run(verbose=True)
    assert np.isclose(log_z_true, sampler.log_z, rtol=0, atol=0.1)
    # Check whether the boundaries nautilus drew are strictly nested.
    shell_bound_occupation = sampler.shell_bound_occupation()
    if np.all(shell_bound_occupation ==
              np.tril(np.ones_like(shell_bound_occupation))):
        warnings.warn('The funnel distribution was too easy.', RuntimeWarning)


def test_sampler_constant():
    # Test that the sampler handles a single constant likelihood gracefully.

    def prior(x):
        return x

    def likelihood(x):
        return 0

    sampler = Sampler(prior, likelihood, 2, n_live=500, seed=0)
    sampler.run(verbose=True, f_live=0.1, n_eff=0)

    assert np.isclose(sampler.log_z, 0)
    # The sampler should not have build any boundaries.
    assert len(sampler.bounds) == 1


def test_sampler_plateau_1():
    # Test that the sampler can deal with a pleateau.

    def prior(x):
        return x

    def likelihood(x):
        if x[0] < 0.9:
            return -np.inf
        else:
            return np.log(x[0] - 0.9)

    log_z_true = np.log(0.5 * 0.1**2)

    for i in range(10):
        sampler = Sampler(prior, likelihood, 2, n_live=1000, n_networks=1,
                          seed=i)
        sampler.run(verbose=True, f_live=0.1)
        assert np.isclose(sampler.log_z, log_z_true, rtol=0, atol=0.1)


def test_sampler_plateau_2():
    # Test that the sampler can deal with a pleateau.

    def prior(x):
        return x

    def likelihood(x):
        return np.ceil(-np.log10(1 - x[0]))

    log_z_true = np.log(
        np.sum(0.9 * 0.1**np.arange(100) * np.exp(1 + np.arange(100))))

    for i in range(10):
        sampler = Sampler(prior, likelihood, 2,
                          n_live=2000, n_networks=1, seed=0)
        sampler.run(verbose=True, f_live=1e-6)
        assert np.isclose(sampler.log_z, log_z_true, atol=0.1)
        # The minimum likelihood values should correspond to the different
        # plateaus.
        assert np.all(np.isclose(sampler.shell_log_l_min[1:],
                      np.arange(len(sampler.bounds) - 1) + 2, rtol=0))



"""Module implementing helper functions for working with pools."""

from multiprocessing import Pool


def initialize_worker(likelihood):
    """
    Initialize a worker for likelihood evaluations.

    Parameters
    ----------
    likelihood : function
        Likelihood function that each worker will evaluate.
    """
    global LIKELIHOOD
    LIKELIHOOD = likelihood


def likelihood_worker(*args):
    """
    Have the worker evaluate the likelihood.

    Parameters
    ----------
    *args : tuple
        Arguments to be passed to the likelihood function.

    Returns
    -------
    object
        Return value of the likelihood function.
    """
    return LIKELIHOOD(*args)


class NautilusPool:
    """
    Wrapper for avoiding implementation-specific details elsewhere.

    Attributes
    ----------
    pool : object or None
        Pool used for parallelization.

    """

    def __init__(self, pool, likelihood=None):
        """
        Initialize a pool.

        Parameters
        ----------
        pool : object
            Pool used for parallelization. If a number, initialize a pool
            from the `multiprocessing` library with the specified number of
            workers.

        """
        if isinstance(pool, int):
            self.pool = Pool(pool, initializer=initialize_worker,
                             initargs=(likelihood, ))
        else:
            self.pool = pool

    def map(self, func, iterable):
        """
        Loop a function over an iterable, similar to the built-in map function.

        Parameters
        ----------
        func : function
            Function to use.
        iterable : object
            Iterable object such as a list or numpy array.

        Returns
        -------
        result : list
            Result of applying the function to all values in the iterable.

        """
        if 'distributed.client.Client' in str(type(self.pool)):
            return list(self.pool.gather(self.pool.map(func, iterable)))
        else:
            return list(self.pool.map(func, iterable))

    @property
    def size(self):
        """Determine the number of workers in the pool.

        Returns
        -------
        size : int
            Size of the pool.

        Raises
        ------
        ValueError
            If the pool size cannot be determined.


        """
        if 'distributed.client.Client' in str(type(self.pool)):
            return len(self.pool.nthreads())
        for attr in ['_processes', '_max_workers', 'size', 'nt']:
            if hasattr(self.pool, attr):
                return getattr(self.pool, attr)
        raise ValueError('Cannot determine size of pool.')



import numpy as np
import multiprocessing
import os
import pytest
import time

from multiprocessing import Pool
from dask.distributed import Client

from nautilus import Sampler
from nautilus.pool import NautilusPool


def prior(x):
    return x


def likelihood(x):
    time.sleep(0.001)
    return 1, os.getpid()


@pytest.mark.skipif(multiprocessing.get_start_method() == 'spawn',
                    reason=('pytest does not support spawning'))
@pytest.mark.parametrize("pool", [None, 1, 2, (1, 4), (4, 1), 'mp', 'dask'])
def test_pool(pool):
    # Test that the expected number of processes are run.

    if isinstance(pool, tuple):
        n_jobs = pool[0]
    else:
        if pool not in ['mp', 'dask']:
            n_jobs = 1 if pool is None else pool
        elif pool == 'mp':
            n_jobs = 4
            pool = Pool(n_jobs)
        else:
            pool = Client()
            n_jobs = NautilusPool(pool).size

    sampler = Sampler(prior, likelihood, n_dim=2, n_live=50, n_networks=1,
                      pool=pool)
    sampler.run(f_live=1.0, n_eff=0)
    points, log_w, log_l, blobs = sampler.posterior(return_blobs=True)

    assert len(np.unique(blobs)) == n_jobs
    assert sampler.n_batch >= 100
    assert (sampler.n_batch % n_jobs) == 0

    try:
        pool.close()
    except AttributeError:
        pass
